import weakref
import TDStoreTools

def createProperty(classInstance, name, value=None, attributeName=None,
						 readOnly=False, dependable=True):
	"""
	Use this method to add a property (called name) that accesses
	an attribute (called attributeName). attributeName defaults to '_' + name.
	The attribute will be set to value argument.
	If dependable is True, the attribute will store value in a dependency obj.
		If dependable is 'deep', lists, sets, or dicts will be created as
		dependable collections from TDStoreTools
	If readonly is True, the property will be read only
	WARNING: the property is added to the CLASS of instance, so all objects of
	instance's class will now have this property.
	"""
	if dependable == 'Deep':
		dependable = 'deep'
	if attributeName is None:
		attributeName = '_' + name
	classInstance = weakref.proxy(classInstance)
	if dependable:
		if dependable == 'deep':
			depValue = makeDeepDependable(value)
		else:
			depValue = tdu.Dependency(value)
		setattr(classInstance, attributeName, depValue)
		def getter(self):
			if dependable == 'deep' and readOnly:
				try:
					return getattr(self, attributeName).val.getRaw()
				except:
					return getattr(self, attributeName).val
			else:
				try:
					return getattr(self, attributeName).val
				except:
					raise Exception(
							f'Unable to access {name} in {classInstance}.\n'
							f'Possible duplicate name.')
		def setter(self, val):
			getattr(self, attributeName).val = val
		def deleter(self):
			getattr(self, attributeName).modified()
			delattr(self, attributeName)
	else:
		setattr(classInstance, attributeName, value)
		def getter(self):
			return getattr(self, attributeName)
		def setter(self, val):
			setattr(self, attributeName, val)
		def deleter(self):
			getattr(self, attributeName).modified()
			delattr(self, attributeName)
	setattr(classInstance.__class__, name, property(getter,
											   None if readOnly else setter,
											   deleter))

def makeDeepDependable(value):
	"""
	returns a deeply dependable object out of the provided python object.
	Deeply dependable collections will cause cooks when their contents change.
	"""
	if isinstance(value, dict):
		depValue = TDStoreTools.DependDict(value)
	elif isinstance(value, list):
		depValue = TDStoreTools.DependList(value)
	elif isinstance(value, set):
		depValue = TDStoreTools.DependSet(value)
	else:
		depValue = tdu.Dependency(value)
	return depValue